/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

/* eslint-disable camelcase */
const event = require('../event')

const users = []
const util = require('../util')
const short = require('short-uuid')
const userService = require('../soap/services/user')

class User {
  constructor (opt) {
    this.userdata = opt.user
    this.context = opt.context
    return new Proxy(this, {
      get (target, prop) {
        return prop in target ? target[prop] : target.userdata[prop]
      }
    })
  }

  async remove () {
    await userService.remove(this.context, this.get('id'))
    users.splice(0, users.length, ...users.filter(u => u !== this))
    event.emit(event.provisioning.user.removed, this)
  }

  async hasConfig (key, value) {
    // Config structure is the following { userAttributes: { entries: [{ key: 'config', value: { entries: [{ key: 'key', value: value }] }}]} }
    // Note that the soap lib will replace arrays with a single element by that element

    // check if config exists and create config path if it does not exist
    const userAttributes = this.userdata.userAttributes = this.userdata.userAttributes || {}
    const entries = userAttributes.entries = userAttributes.entries ? [].concat(userAttributes.entries) : []
    let configEntry = entries.find(entry => entry.key === 'config')
    if (!configEntry) {
      entries.push({ key: 'config', value: {} })
      configEntry = entries[entries.length - 1]
    }
    const config = configEntry.value
    const configEntries = config.entries = config.entries ? [].concat(config.entries) : []
    let targetConfig = configEntries.find(entry => entry.key === key)
    if (!targetConfig) {
      configEntries.push({ key })
      targetConfig = configEntries[configEntries.length - 1]
    }
    targetConfig.value = value

    return userService.change(this.context, this.userdata)
  }

  hasAlias (alias) {
    // See if this alias has already been declared
    if (this.userdata.aliases.includes(alias)) return
    // If not, add it and save
    this.userdata.aliases.push(alias)
    return userService.change(this.context, this.userdata)
  }

  doesntHaveAlias (alias) {
    const index = this.userdata.aliases.indexOf(alias)
    // See if this alias has already been declared
    if (index < 0) return
    this.userdata.aliases.splice(index)
    return userService.change(this.context, this.userdata)
  }

  hasAccessCombination (accessCombinationName) {
    return userService.changeByModuleAccessName(this.context, this.get('id'), accessCombinationName)
  }

  getModuleAccess () {
    return userService.getModuleAccess(this.context, this.get('id'))
  }

  async hasModuleAccess (moduleAccess) {
    const currentAccess = await this.getModuleAccess()
    return userService.changeByModuleAccess(this.context, this.get('id'), currentAccess, moduleAccess)
  }

  hasCapability (capsToAdd) {
    return userService.changeCapabilities(this.context, this.get('id'), capsToAdd)
  }

  doesntHaveCapability (capsToRemove) {
    return userService.changeCapabilities(this.context, this.get('id'), undefined, capsToRemove)
  }

  get login () {
    return `${this.get('name')}@${this.context.id}`
  }

  get (key) {
    return this.userdata[key]
  }

  toJSON () {
    return {
      context: { ...this.context },
      ...this.userdata
    }
  }

  static getRandom ({ name = 'test.user', password = util.getDefaultUserPassword(), sur_name = '', given_name = 'User' } = {}) {
    const id = short.generate().slice(0, 9).toLowerCase()
    if (!sur_name) sur_name = id
    if (name === 'test.user') name = `${name}-${id}`
    const domain = util.mxDomain()
    return {
      name,
      primaryEmail: `${name}@${domain}`,
      email1: `${name}@${domain}`,
      password,
      sur_name,
      given_name
    }
  }

  static async create (user = this.getRandom(), ctx = {}) {
    const usrdata = Object.assign({
      display_name: user.name,
      imapLogin: user.name,
      imapServer: util.imapServer(),
      smtpServer: util.smtpServer()
    }, user)
    event.emit(event.provisioning.user.create, usrdata, ctx)
    const contexts = require('../contexts/contexts')()
    let userContext
    try {
      // @ts-ignore
      userContext = await contexts.reuse(ctx, ctx.admin, ctx.auth)
    } catch (e) {
      // @ts-ignore
      userContext = await contexts.create(Object.assign({ maxQuota: 1000 }, ctx))
    }
    const data = await userService.create(userContext, usrdata)
    const index = users.push(new User({ user: data, context: userContext })) - 1
    event.emit(event.provisioning.user.created, users[index])
    return users[index]
  }

  change (userdata) {
    return userService.change(this.context, userdata)
  }

  static async removeAll () {
    let user = users.pop()
    while (user) {
      await user.remove()
      user = users.pop()
    }
    return users
  }
}

module.exports = () => {
  return new Proxy(users, {
    // act like an array if an index is being used
    get (target, prop) {
      if (prop in target) { return target[prop] }
      return User[prop]
    }
  })
}
