import { describe, it, expect, afterEach, jest } from '@jest/globals'
import vitePluginProxy from '../../dist/index'
import { closeAllServers, createTestServer, createViteServer, filterPingCalls, generateSelfsigned, getPort, getDirname, http2Request, updateRequestManager } from '../util'
import path from 'node:path'

const PORT = getPort()
const SERVER_PORT = getPort()
const SERVER_PORT2 = getPort()
const __dirname = getDirname(import.meta.url)

describe('enhanced frontend configuration', function () {
  afterEach(async function () {
    await closeAllServers()
    updateRequestManager()
  })

  it('select a preferred frontend by path', async function () {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT}`,
          path: '/core'
        }, {
          target: `http://localhost:${SERVER_PORT2}`,
          path: '/plugin'
        }]
      })]
    }, PORT)

    const server = createTestServer()
    const mockCallback = jest.fn((req, res) => res.send('alert("foo")'))
    server.get('*', mockCallback)
    await server.listen({ port: SERVER_PORT })

    const server2 = createTestServer()
    const mockCallback2 = jest.fn((req, res) => res.send('alert("bar")'))
    server2.get('*', mockCallback2)
    await server2.listen({ port: SERVER_PORT2 })

    expect(filterPingCalls(mockCallback)).toBeCalledTimes(0)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(0)

    const { data } = await http2Request(`http://localhost:${PORT}/core/file.js`)
    expect(data).toEqual('alert("foo")')
    expect(filterPingCalls(mockCallback)).toBeCalledTimes(1)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(0)

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/plugin/file.js`)
    expect(data2).toEqual('alert("bar")')
    expect(filterPingCalls(mockCallback)).toBeCalledTimes(1)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(1)
  })

  it('select a preferred frontend by regex', async function () {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT}`,
          path: '^/core'
        }, {
          target: `http://localhost:${SERVER_PORT2}`,
          path: '^/(?!core)'
        }]
      })]
    }, PORT)

    const server = createTestServer()
    const mockCallback = jest.fn((req, res) => res.send('alert("foo")'))
    server.get('*', mockCallback)
    await server.listen({ port: SERVER_PORT })

    const server2 = createTestServer()
    const mockCallback2 = jest.fn((req, res) => res.send('alert("bar")'))
    server2.get('*', mockCallback2)
    await server2.listen({ port: SERVER_PORT2 })

    expect(filterPingCalls(mockCallback)).toBeCalledTimes(0)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(0)

    const { data } = await http2Request(`http://localhost:${PORT}/core/file.js`)
    expect(data).toEqual('alert("foo")')
    expect(filterPingCalls(mockCallback)).toBeCalledTimes(1)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(0)

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/plugin/file.js`)
    expect(data2).toEqual('alert("bar")')
    expect(filterPingCalls(mockCallback)).toBeCalledTimes(1)
    expect(filterPingCalls(mockCallback2)).toBeCalledTimes(1)
  })

  it('supports secure=false', async function () {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `https://localhost:${SERVER_PORT}`,
          secure: false
        }]
      })]
    }, PORT)

    const server = createTestServer({
      http2: true,
      https: generateSelfsigned()
    })
    server.get('*', (req, res) => res.send('alert("foo")'))
    await server.listen({ port: SERVER_PORT })

    await http2Request(`http://localhost:${PORT}/plugin/file.js`)
  })

  it('supports frontends with different base configurations', async function () {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: path.join(__dirname, 'appsuite'),
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy()]
    }, SERVER_PORT2)
    const { data } = await http2Request(`http://localhost:${PORT}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'foo\'\n'))

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/bar.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'bar\'\n'))
  })

  it('supports frontends with different base configurations (2)', async function () {
    await createViteServer({
      root: path.join(__dirname, 'appsuite'),
      logLevel: 'silent',
      base: '/appsuite/',
      server: {
        fs: {
          allow: ['/']
        }
      },
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/',
      server: {
        fs: {
          allow: ['/']
        }
      },
      plugins: [vitePluginProxy()]
    }, SERVER_PORT2)

    const { data } = await http2Request(`http://localhost:${PORT}/appsuite/bar.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'bar\'\n'))

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/appsuite/index.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'foo\'\n'))
  })

  it('supports frontends with different base configurations (3)', async function () {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}/appsuite/`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: path.join(__dirname, 'appsuite'),
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy()]
    }, SERVER_PORT2)

    const { data } = await http2Request(`http://localhost:${PORT}/appsuite/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'foo\'\n'))

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/appsuite/bar.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'bar\'\n'))
  })

  it('proxies requests to frontends that allow file system access', async () => {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/',
      plugins: [vitePluginProxy()]
    }, PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/@fs${__dirname}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'foo\'\n'))
    const { data: data2 } = await http2Request(`http://localhost:${PORT}/@fs${__dirname}/appsuite/bar.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'bar\'\n'))
  })

  it('proxies requests to frontends with base configured that allow file system access', async () => {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}/appsuite/`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy()]
    }, PORT)
    const { data } = await http2Request(`http://localhost:${PORT}/appsuite/@fs${__dirname}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'foo\'\n'))
    const { data: data2 } = await http2Request(`http://localhost:${PORT}/appsuite/@fs${__dirname}/appsuite/bar.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'bar\'\n'))
  })

  it('proxies requests to frontends with different bases that allow file system access', async () => {
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/',
      plugins: [vitePluginProxy({
        frontends: [{
          target: `http://localhost:${SERVER_PORT2}`
        }]
      })]
    }, PORT)
    await createViteServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy({})]
    }, PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/@fs${__dirname}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toEqual(expect.stringContaining('window.test = \'foo\'\n'))
    const { data: data2 } = await http2Request(`http://localhost:${PORT}/@fs${__dirname}/appsuite/bar.js`)
    expect(data2).toEqual(expect.stringContaining('window.test = \'bar\'\n'))
  })
})
